// $Id: LOCA_MultiContinuation_CompositeConstraintMVDX.C,v 1.1 2005/06/23 22:53:55 etphipp Exp $
// $Source: /space/CVS/Trilinos/packages/nox/src-loca/src/LOCA_MultiContinuation_CompositeConstraintMVDX.C,v $

//@HEADER
// ************************************************************************
//
//                  LOCA Continuation Algorithm Package
//                 Copyright (2005) Sandia Corporation
//
// Under terms of Contract DE-AC04-94AL85000, there is a non-exclusive
// license for use of this work by or on behalf of the U.S. Government.
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
// Questions? Contact Andy Salinger (agsalin@sandia.gov) or Eric Phipps
// (etphipp@sandia.gov), Sandia National Laboratories.
//
// ************************************************************************
//@HEADER

#include "LOCA_MultiContinuation_CompositeConstraintMVDX.H"
#include "LOCA_GlobalData.H"
#include "LOCA_ErrorCheck.H"

LOCA::MultiContinuation::CompositeConstraintMVDX::CompositeConstraintMVDX(
    const Teuchos::RefCountPtr<LOCA::GlobalData>& global_data,
    const vector< Teuchos::RefCountPtr<LOCA::MultiContinuation::ConstraintInterfaceMVDX> >& constraintObjects) :
  LOCA::MultiContinuation::CompositeConstraint(),
  constraintMVDXPtrs(constraintObjects),
  compositeDX()
{
  // Copy constraint object pointers into temporary array of base class
  vector<Teuchos::RefCountPtr<LOCA::MultiContinuation::ConstraintInterface> > tmp(constraintObjects.size());
  for (unsigned int i=0; i<constraintObjects.size(); i++)
    tmp[i] = constraintObjects[i];

  // Initialize parent class
  init(global_data, tmp);

  // Find first constraint that has nonzero DX
  int i=0;
  while (i < numConstraintObjects && constraintPtrs[i]->isDXZero())
    i++;

  // Create new multivector to store all constraint derivatives
  if (i < numConstraintObjects)
    compositeDX = 
      Teuchos::rcp(constraintMVDXPtrs[i]->getDX()->clone(totalNumConstraints));
  else
    compositeDX = Teuchos::null;
}

LOCA::MultiContinuation::CompositeConstraintMVDX::CompositeConstraintMVDX(
	      const LOCA::MultiContinuation::CompositeConstraintMVDX& source, 
	      NOX::CopyType type) : 
  LOCA::MultiContinuation::CompositeConstraint(source),
  compositeDX()
{
  if (source.compositeDX.get() != NULL)
    compositeDX = Teuchos::rcp(source.compositeDX->clone(type));
  else
    compositeDX = Teuchos::null;
}

LOCA::MultiContinuation::CompositeConstraintMVDX::~CompositeConstraintMVDX()
{
}

LOCA::MultiContinuation::CompositeConstraintMVDX&
LOCA::MultiContinuation::CompositeConstraintMVDX::operator=(
		   const LOCA::MultiContinuation::CompositeConstraintMVDX& source)
{
  if (this != &source) {
    LOCA::MultiContinuation::CompositeConstraint::operator=(source);
    if (compositeDX.get() != NULL && source.compositeDX.get() != NULL)
      *compositeDX = *source.compositeDX;
    else if (source.compositeDX.get() != NULL)
      compositeDX = Teuchos::rcp(source.compositeDX->clone(NOX::DeepCopy));
    else
      compositeDX = Teuchos::null;
  }

  return *this;
}

LOCA::MultiContinuation::ConstraintInterface& 
LOCA::MultiContinuation::CompositeConstraintMVDX::operator=(
		   const LOCA::MultiContinuation::ConstraintInterface& source)
{
  return operator=(dynamic_cast<const LOCA::MultiContinuation::CompositeConstraintMVDX&>(source));
}

Teuchos::RefCountPtr<LOCA::MultiContinuation::ConstraintInterface>
LOCA::MultiContinuation::CompositeConstraintMVDX::clone(NOX::CopyType type) const
{
  return Teuchos::rcp(new CompositeConstraintMVDX(*this, type));
}

NOX::Abstract::Group::ReturnType
LOCA::MultiContinuation::CompositeConstraintMVDX::computeDX()
{
    string callingFunction = 
    "LOCA::MultiContinuation::CompositeConstraintMVDX::computeConstraints()";
  NOX::Abstract::Group::ReturnType status;
  NOX::Abstract::Group::ReturnType finalStatus = NOX::Abstract::Group::Ok;

  if (isValidDX)
    return finalStatus;

  if (isDXZero())
    return finalStatus;

  Teuchos::RefCountPtr<NOX::Abstract::MultiVector> dx;
  for (int i=0; i<numConstraintObjects; i++) {

    if (!constraintMVDXPtrs[i]->isDXZero()) {

      // Compute constraint derivative
      status = constraintMVDXPtrs[i]->computeDX();
      finalStatus = 
	globalData->locaErrorCheck->combineAndCheckReturnTypes(
							     status, 
							     finalStatus,
							     callingFunction);
      
      // Copy columns of constraint dervative into composite
      dx = Teuchos::rcp(compositeDX->subView(indices[i]));
      *dx = *(constraintMVDXPtrs[i]->getDX());
    }
  }

  return finalStatus;
}

NOX::Abstract::Group::ReturnType
LOCA::MultiContinuation::CompositeConstraintMVDX::multiplyDX(
		      double alpha, 
		      const NOX::Abstract::MultiVector& input_x,
	              NOX::Abstract::MultiVector::DenseMatrix& result_p) const
{
  return 
    LOCA::MultiContinuation::ConstraintInterfaceMVDX::multiplyDX(alpha,
								 input_x,
								 result_p);
}

NOX::Abstract::Group::ReturnType
LOCA::MultiContinuation::CompositeConstraintMVDX::addDX(
		              Teuchos::ETransp transb,
			      double alpha, 
	                      const NOX::Abstract::MultiVector::DenseMatrix& b,
			      double beta,
			      NOX::Abstract::MultiVector& result_x) const
{
  return 
    LOCA::MultiContinuation::ConstraintInterfaceMVDX::addDX(transb, alpha, b,
							    beta, result_x);
}

const NOX::Abstract::MultiVector*
LOCA::MultiContinuation::CompositeConstraintMVDX::getDX() const
{
  return compositeDX.get();
}
